@@ -51,7 +51,6 @@
 #include <libpanel-util/panel-keyfile.h>
 #include <libpanel-util/panel-show.h>
 
-#include "nothing.h"
 #include "panel-util.h"
 #include "panel-globals.h"
 #include "panel-enums.h"
@@ -115,25 +114,33 @@ static void panel_run_dialog_disconnect_pixmap (PanelRunDialog *dialog);
 
 #define PANEL_RUN_SCHEMA "org.mate.panel"
 #define PANEL_RUN_HISTORY_KEY "history-mate-run"
+#define PANEL_RUN_HISTORY_MAX_SIZE_KEY "history-max-size-mate-run"
+#define PANEL_RUN_HISTORY_REVERSE_KEY "history-reverse-mate-run"
 #define PANEL_RUN_SHOW_PROGRAM_LIST_KEY "show-program-list"
-#define PANEL_RUN_MAX_HISTORY 10
 
 static GtkTreeModel *
 _panel_run_get_recent_programs_list (PanelRunDialog *dialog)
 {
 	GtkListStore *list;
 	gchar       **items;
-	int           i = 0;
+	guint         history_max_size;
+	gboolean      history_reverse = FALSE;
+	guint         i = 0;
 
 	list = gtk_list_store_new (1, G_TYPE_STRING);
 
+	history_max_size = g_settings_get_uint (dialog->settings, PANEL_RUN_HISTORY_MAX_SIZE_KEY);
+	history_reverse = g_settings_get_boolean (dialog->settings, PANEL_RUN_HISTORY_REVERSE_KEY);
 	items = g_settings_get_strv (dialog->settings, PANEL_RUN_HISTORY_KEY);
-
 	for (i = 0;
-	     items[i] && i < PANEL_RUN_MAX_HISTORY;
+	     items[i] && i < history_max_size;
 	     i++) {
 		GtkTreeIter iter;
-		gtk_list_store_prepend (list, &iter);
+		/* add history in reverse */
+		if (history_reverse)
+			gtk_list_store_prepend (list, &iter);
+		else
+			gtk_list_store_append (list, &iter);
 		gtk_list_store_set (list, &iter, 0, items[i], -1);
 	}
 
@@ -149,31 +156,46 @@ _panel_run_save_recent_programs_list (PanelRunDialog   *dialog,
 {
 	GtkTreeModel *model;
 	GtkTreeIter   iter;
-	GArray       *items;
-	gint          i = 0;
+	guint         history_max_size;
+	gboolean      history_reverse;
 
-	items = g_array_new (TRUE, TRUE, sizeof (gchar *));
-	g_array_append_val (items, lastcommand);
-	i++;
+	history_reverse = g_settings_get_boolean (dialog->settings, PANEL_RUN_HISTORY_REVERSE_KEY);
+	history_max_size = g_settings_get_uint (dialog->settings, PANEL_RUN_HISTORY_MAX_SIZE_KEY);
 
-	model = gtk_combo_box_get_model (GTK_COMBO_BOX (entry));
+	if (history_max_size == 0)
+		g_settings_set_strv (dialog->settings, PANEL_RUN_HISTORY_KEY, NULL);
+	else {
+		model = gtk_combo_box_get_model (GTK_COMBO_BOX (entry));
 
-	if (gtk_tree_model_get_iter_first (model, &iter)) {
-		char *command;
-		do {
-			gtk_tree_model_get (model, &iter, 0, &command, -1);
-			if (g_strcmp0 (command, lastcommand) == 0)
-				continue;
-			g_array_append_val (items, command);
-			i++;
-		} while (gtk_tree_model_iter_next (model, &iter) &&
-			 i < PANEL_RUN_MAX_HISTORY);
+		/* reasonable upper bound for zero-terminated array with new command */
+		gchar const *items[gtk_tree_model_iter_n_children (model, NULL) + 2];
+		guint        items_added = 0;
+
+		items[0] = lastcommand;
+
+		if (gtk_tree_model_get_iter_first (model, &iter)) {
+			char *command;
+			do {
+				gtk_tree_model_get (model, &iter, 0, &command, -1);
+				if (g_strcmp0 (command, lastcommand) == 0)
+					continue;
+				items[items_added + 1] = command;
+				items_added++;
+			} while (gtk_tree_model_iter_next (model, &iter));
+		}
+
+		if (history_reverse) {
+			guint pos;
+			for (pos = 0; pos < items_added / 2; pos++)
+			{
+				gchar const * tmp = items[pos + 1];
+				items[pos + 1] = items[items_added - pos];
+				items[items_added - pos] = tmp;
+			}
+		}
+		items[history_max_size < items_added + 1 ? history_max_size : items_added + 1] = NULL;
+		g_settings_set_strv (dialog->settings, PANEL_RUN_HISTORY_KEY, items);
 	}
-
-	g_settings_set_strv (dialog->settings, PANEL_RUN_HISTORY_KEY,
-			     (const gchar **) items->data);
-
-	g_array_free (items, TRUE);
 }
 
 static void
@@ -271,12 +293,7 @@ panel_run_dialog_set_icon (PanelRunDialog *dialog,
 		GtkIconTheme *icon_theme = gtk_icon_theme_get_default ();
 		GtkIconInfo *icon_info = gtk_icon_theme_lookup_by_gicon (icon_theme, icon, size, 0);
 		pixbuf = gtk_icon_info_load_icon (icon_info, NULL);
-#if GTK_CHECK_VERSION (3, 8, 0)
 		g_object_unref (icon_info);
-#else
-		gtk_icon_info_free (icon_info);
-#endif
-
 	}
 
 	if (pixbuf) {
@@ -291,11 +308,7 @@ panel_run_dialog_set_icon (PanelRunDialog *dialog,
 		//(ditto for the drag icon?)
 		gtk_window_set_icon (GTK_WINDOW (dialog->run_dialog), pixbuf);
 
-#if GTK_CHECK_VERSION (3, 2, 0)
 		gtk_drag_source_set_icon_gicon (dialog->run_dialog, dialog->icon);
-#else
-		gtk_drag_source_set_icon_pixbuf (dialog->run_dialog, pixbuf);
-#endif
 		g_object_unref (pixbuf);
 	} else {
 		panel_run_dialog_set_default_icon (dialog, TRUE);
@@ -346,6 +359,15 @@ command_is_executable (const char   *command,
 	return TRUE;
 }
 
+/*
+ * Set the DISPLAY variable, to be use by g_spawn_async.
+ */
+static void
+set_environment (gpointer display)
+{
+  g_setenv ("DISPLAY", display, TRUE);
+}
+
 static void
 dummy_child_watch (GPid         pid,
 		   		   gint         status,
@@ -367,6 +389,7 @@ panel_run_dialog_launch_command (PanelRunDialog *dialog,
 	GError     *error = NULL;
 	char      **argv;
 	int         argc;
+	char       *display;
 	GPid        pid;
 
 	if (!command_is_executable (locale_command, &argc, &argv))
@@ -377,26 +400,16 @@ panel_run_dialog_launch_command (PanelRunDialog *dialog,
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->terminal_checkbox)))
 		mate_desktop_prepend_terminal_to_vector (&argc, &argv);
 
-#if GTK_CHECK_VERSION (3, 0, 0)
+	display = gdk_screen_make_display_name (screen);
+
 	result = g_spawn_async (NULL, /* working directory */
 				argv,
 				NULL, /* envp */
-				G_SPAWN_SEARCH_PATH,
-				NULL,
-				NULL,
+				G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
+				set_environment,
+				display,
 				&pid,
 				&error);
-#else
-	result = gdk_spawn_on_screen (screen,
-				      NULL, /* working directory */
-				      argv,
-				      NULL, /* envp */
-				      G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD,
-				      NULL, /* child setup func */
-				      NULL, /* user data */
-				      &pid, /* child pid */
-				      &error);
-#endif
 
 	if (!result) {
 		char *primary;
@@ -435,24 +448,6 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 		return;
 	}
 
-	/* evil eggies, do not translate! */
-#if !GTK_CHECK_VERSION (3, 0, 0)
-	/* FIXME re-add once GTK3 support is fixed */
-	if (!strcmp (command, "free the fish")) {
-		start_screen_check ();
-
-		g_free (command);
-		gtk_widget_destroy (dialog->run_dialog);
-		return;
-	} else if (!strcmp (command, "gegls from outer space")) {
-		start_geginv ();
-
-		g_free (command);
-		gtk_widget_destroy (dialog->run_dialog);
-		return;
-	}
-#endif
-
 	error = NULL;
 	disk = g_locale_from_utf8 (command, -1, NULL, NULL, &error);
 
@@ -1090,6 +1085,8 @@ panel_run_dialog_setup_program_list (PanelRunDialog *dialog,
 	dialog->program_label = PANEL_GTK_BUILDER_GET (gui, "program_label");
 	dialog->main_box = PANEL_GTK_BUILDER_GET (gui, "main_box");
 
+	gtk_widget_set_can_focus (dialog->program_label, FALSE);
+
 	/* Ref the box so it doesn't get destroyed when it is
 	 * removed from the visible area of the dialog box.
 	 */
@@ -1831,14 +1828,8 @@ pixmap_drag_data_get (GtkWidget          *run_dialog,
 }
 
 static void
-#if GTK_CHECK_VERSION (3, 0, 0)
 panel_run_dialog_style_updated (GtkWidget *widget,
 				PanelRunDialog *dialog)
-#else
-panel_run_dialog_style_set (GtkWidget      *widget,
-			    GtkStyle       *prev_style,
-			    PanelRunDialog *dialog)
-#endif
 {
   if (dialog->icon) {
 	  GIcon *icon;
@@ -1869,15 +1860,9 @@ panel_run_dialog_setup_pixmap (PanelRunDialog *dialog,
 {
 	dialog->pixmap = PANEL_GTK_BUILDER_GET (gui, "icon_pixmap");
 
-#if GTK_CHECK_VERSION (3, 0, 0)
 	g_signal_connect (dialog->pixmap, "style-updated",
 			  G_CALLBACK (panel_run_dialog_style_updated),
 			  dialog);
-#else
-	g_signal_connect (dialog->pixmap, "style-set",
-			  G_CALLBACK (panel_run_dialog_style_set),
-			  dialog);
-#endif
 	g_signal_connect (dialog->pixmap, "screen-changed",
 			  G_CALLBACK (panel_run_dialog_screen_changed),
 			  dialog);
@@ -1938,15 +1923,9 @@ panel_run_dialog_new (GdkScreen  *screen,
 static void
 panel_run_dialog_disconnect_pixmap (PanelRunDialog *dialog)
 {
-#if GTK_CHECK_VERSION (3, 0, 0)
 	g_signal_handlers_disconnect_by_func (dialog->pixmap,
 					      G_CALLBACK (panel_run_dialog_style_updated),
 					      dialog);
-#else
-	g_signal_handlers_disconnect_by_func (dialog->pixmap,
-			                      G_CALLBACK (panel_run_dialog_style_set),
-			                      dialog);
-#endif
 	g_signal_handlers_disconnect_by_func (dialog->pixmap,
 			                      G_CALLBACK (panel_run_dialog_screen_changed),
 			                      dialog);
